'''
@brief vimanGitWrapper wrapper of git subprocess
@note the only API to git operations
'''

import subprocess
import os
import sys
import errno

import yaml

from viman import vimanYamlWrapper
from viman import vimanUtils

class vimanGitWrapper():
    '''
    @brief low operation wrapper of git , install , upgrade , remove , query
    @note all other operation base on this
    '''

    DIR_PLUGIN = os.path.join(os.getenv('HOME'), '.vim/bundle')

    @staticmethod
    @vimanYamlWrapper.vimanYamlWrapper.installWrapper
    def install(url, recipe):
        '''
        @brief install a vim plugin to .vim/bundle by git from url
        @param url url of git repository
        '''
        if not os.path.isdir(vimanGitWrapper.DIR_PLUGIN):
            os.makedirs(vimanGitWrapper.DIR_PLUGIN)
        pwd = os.getenv('PWD')
        os.chdir(vimanGitWrapper.DIR_PLUGIN)
        if(hasattr(subprocess, 'run')):
            ret = subprocess.run(['git', 'clone', url]).returncode
        else:
            ret = subprocess.call(['git', 'clone', url])
        if not 0 == ret:
            os.chdir(pwd)
            sys.exit(ret)
        ret = os.system(recipe)
        os.chdir(pwd)
        return ret

    @staticmethod
    def installByYml(ymlName):
        '''
        @brief install plugins by yml
        @param ymlName yml file name
        '''
        if not os.path.isfile(ymlName):
            print(
                'error:don\'t exist file `{}`!'.format(ymlName),
                file=sys.stderr)
            sys.exit(errno.EINVAL)
        with open(ymlName) as f:
            yml = yaml.load(f)
            if yml:
                for v in yml.values():
                    vimanGitWrapper.install(v['url'],v['recipe'])

    @staticmethod
    def installByCurrent(ymlString):
        if ymlString:
            yml = yaml.load(ymlString)
            if yml:
                for v in yml.values():
                    vimanGitWrapper.install(v['url'], v['recipe'])

    @staticmethod
    def upgrade(url):
        '''
        @brief upgrade a vim plugin by git from url
        @param url url of git repository
        '''
        return vimanGitWrapper.upgradeByName(
            vimanUtils.vimanUtils.getPlugin4Url(url))

    @staticmethod
    def upgradeByName(name):
        '''
        @brief upgrade plugin by plugin name
        '''
        path = os.path.join(vimanGitWrapper.DIR_PLUGIN,name)
        if not os.path.isdir(path):
            print(
                'error:don\'t exist directory `{}`!'.format(path),
                file=sys.stderr)
            sys.exit(errno.ENOENT)
        pwd = os.getenv('PWD')
        os.chdir(path)
        if hasattr(subprocess, 'run'):
            ret = subprocess.run(['git', 'pull']).returncode
        else:
            ret = subprocess.call(['git', 'pull'])
        if not 0 == ret:
            os.chdir(pwd)
            sys.exit(ret)
        os.chdir(pwd)
        return ret

    @staticmethod
    def upgradeByYml(ymlName):
        if not os.path.isfile(ymlName):
            print(
                'error:don\'t exist file `{}`!'.format(ymlName),
                file=sys.stderr)
            sys.exit(errno.EINVAL)
        with open(ymlName) as f:
            yml = yaml.load(f)
            if yml:
                for v in yml.values():
                    vimanGitWrapper.upgrade(v['url'])

    @staticmethod
    def remove(url):
        '''
        @brief remove a vim plugin from url
        @param url url of git repository
        '''
        return vimanGitWrapper.removeByName(
            vimanUtils.vimanUtils.getPlugin4Url(url))

    @staticmethod
    @vimanYamlWrapper.vimanYamlWrapper.removeByNameWrapper
    def removeByName(name):
        path = os.path.join(vimanGitWrapper.DIR_PLUGIN, name)
        if not os.path.isdir(path):
            print(
                'error:don\'t exist directory `{}`!'.format(path),
                file=sys.stderr)
            sys.exit(errno.ENOENT)
        if hasattr(subprocess, 'run'):
            ret = subprocess.run(['rm', '-rf', path]).returncode
        else:
            ret =  subprocess.call(['rm', '-rf', path])
        if not 0 == ret:
            sys.exit(ret)
        return ret

    @staticmethod
    def removeByYml(ymlName):
        '''
        @brief remove plugins by yml
        @param ymlName yml file name
        '''
        if not os.path.isfile(ymlName):
            print(
                'error:don\'t exist file `{}`!'.format(ymlName),
                file=sys.stderr)
            sys.exit(errno.EINVAL)
        with open(ymlName) as f:
            yml = yaml.load(f)
            if yml:
                for v in yml.values():
                    vimanGitWrapper.remove(v['url'])




#! /usr/bin/env python3

'''
@biref viman cli entry
'''

import sys
import errno

from viman import vimanArgParser
from viman import vimanGitWrapper
from viman import vimanYamlWrapper
from viman.__init__ import __version__
from viman.__init__ import __program__

PROGRAM = __program__
VERSION = __version__

# errno = ['OK',                          #0
# '~/.vim/bundle don\'t exists!'] #1


def main():
    '''
    @brief viman for vim plugins manager
    @return error
    '''

    parser = vimanArgParser.vimanArgParser(sys.argv[1:])
    # print(parser.operations)
    # print(parser.options)
    # print(parser.targets)

    if parser.operations[0] == vimanArgParser.vimanOperations.\
            operations['sync'][0]:
        # sync
        if vimanArgParser.vimanOptions.options['file'][0] in parser.options:
            # yml file
            for yml_file in parser.targets:
                vimanGitWrapper.vimanGitWrapper.installByYml(yml_file)
        elif vimanArgParser.vimanOptions.options['current'][0] in \
                parser.options:
            # current yml string
            for yml in parser.targets:
                vimanGitWrapper.vimanGitWrapper.installByCurrent(yml)
        elif vimanArgParser.vimanOptions.options['sysupgrade'][0] in \
                parser.options:
            # sysupgrade
            vimanGitWrapper.vimanGitWrapper.upgradeByYml(
                vimanYamlWrapper.vimanYamlWrapper.ymlDefault)

        else:
            for target in parser.targets:
                vimanGitWrapper.vimanGitWrapper.install(target, '')

    elif parser.operations[0] == vimanArgParser.vimanOperations.\
            operations['remove'][0]:
        if vimanArgParser.vimanOptions.options['file'][0] in parser.options:
            # yml file
            for yml_file in parser.targets:
                vimanGitWrapper.vimanGitWrapper.removeByYml(yml_file)
        elif vimanArgParser.vimanOptions.options['name'][0] in parser.options:
            for name in parser.targets:
                vimanGitWrapper.vimanGitWrapper.removeByName(name)

        else:
            for target in parser.targets:
                vimanGitWrapper.vimanGitWrapper.remove(target)

    elif parser.operations[0] == vimanArgParser.vimanOperations.\
            operations['upgrade'][0]:
        if vimanArgParser.vimanOptions.options['file'][0] in parser.options:
            # upgrade by yml file
            for yml_file in parser.targets:
                vimanGitWrapper.vimanGitWrapper.upgradeByYml(yml_file)
        elif vimanArgParser.vimanOptions.options['recursive'][0] in \
                parser.options:
            # upgrade recursive
            vimanGitWrapper.vimanGitWrapper.upgradeByYml(
                vimanYamlWrapper.vimanYamlWrapper.ymlDefault)
        for target in parser.targets:
            vimanGitWrapper.vimanGitWrapper.upgrade(target)
    elif parser.operations[0] == vimanArgParser.vimanOperations.\
            operations['query'][0]:
        yml = vimanYamlWrapper.vimanYamlWrapper.loadYml()
        if yml is None:
            print(yml)
        else:
            for name in yml:
                print(name)
    elif parser.operations[0] == vimanArgParser.vimanOperations.\
            operations['version'][0]:
        print('{}-{}'.format(PROGRAM, VERSION))
    elif parser.operations[0] == vimanArgParser.vimanOperations.\
            operations['help'][0]:
        # help
        vimanArgParser.vimanArgParser.printUsage()
    else:
        # invalid operation
        print('error:invalid operation `{}`!'.format(parser.operations[0]),
              file=sys.stderr)
        return errno.EINVAL

    return 0

if '__main__' == __name__:
    sys.exit(main())



#! /usr/bin/env python3

'''
@brief vimanYamlWrapper wrapper fo yaml operations
@note not the only entry of yaml API
'''

import os
import sys

import yaml

from viman import vimanUtils

class vimanYamlWrapper():
    '''
    @brief vimanYamlWrapper wrapper of yaml operations
    '''

    ymlDefault = os.path.join(os.getenv('HOME'), '.viman.yml')

    @staticmethod
    def installWrapper(install):
        def wrapper(url, recipe):
            ret = install(url, recipe)  # install plugin
            yml = vimanYamlWrapper.loadYml()
            if yml is None:
                yml = {}
            yml[vimanUtils.vimanUtils.getPlugin4Url(url)] = {
                'url':url, 'recipe':recipe}
            vimanYamlWrapper.dumpYml(yml)
            return ret

        return wrapper

    @staticmethod
    def removeWrapper(remove):
        def wrapper(url):
            ret = remove(url)  # remove plugin
            yml = vimanYamlWrapper.loadYml()
            if yml is None:
                return errno.EINVAL
            try:
                yml.pop(vimanUtils.vimanUtils.getPlugin4Url(url))
            except KeyError:
                print(
                    'error:yml no key `{}`!'.format(
                        vimanUtils.vimanUtils.getPlugin4Url(url)),
                    file=sys.stderr)
            vimanYamlWrapper.dumpYml(yml)
            return ret
        
        return wrapper

    @staticmethod
    def removeByNameWrapper(removeByName):
        def wrapper(name):
            ret = removeByName(name)  # remove plugin by name
            yml = vimanYamlWrapper.loadYml()
            try:
                yml.pop(name)
            except KeyError:
                print('error:yml no key `{}`!'.format(name), file=sys.stderr)

            vimanYamlWrapper.dumpYml(yml)
            return ret

        return wrapper

    @staticmethod
    def loadYml(file=ymlDefault):
        '''
        @brief load yaml object from yaml file
        @param file yaml file name string
        @retval yaml object
        '''
        if not os.path.isfile(file):
            os.mknod(file)
        with open(file, 'r') as f:
            yml = yaml.load(f)
            f.close()
            return yml

    @staticmethod
    def dumpYml(yml, file=ymlDefault):
        with open(file, 'w') as f:
            yaml.dump(yml, stream=f, default_flow_style=False)
            f.close()


def _test():
    with open(vimanYamlWrapper.ymlDefault, 'r') as f:
        # yaml.dump(yaml.load(f),stream = f, default_flow_style=False)
        yml = yaml.load(f)
        print(yml)

if '__main__' == __name__:
    _test()



'''
@brief vimanUtils common utils for viman
'''

import os

class vimanUtils():
    @staticmethod
    def getPlugin4Url(url):
        '''
        @brief get name of plugin from url
        @param url git repository url
        '''
        basename = os.path.basename(url)
        name = os.path.splitext(basename)
        return name[0]



'''
@brief setup for packing
'''

import re

from setuptools import setup


def readme():
    '''
    @brief read content from README
    @retval content string of README
    '''
    with open('README.rst') as f:
        return f.read()


LONG_DESCRIPTION = open('README.rst').read()

def find_attr(attr, file_path):
    '''
    @brief find attribute from documents by regex
    @param attr attribute name string
    @param file path string
    @retval attribute value
    '''
    with open(file_path) as f:
        attr_file = f.read()
    attr_match = re.search(
        r"^{}\s*=\s*['\"]([^'\"]*)['\"]".format(attr),
        attr_file,
        re.M)

    if attr_match:
        return attr_match.group(1)

    raise RuntimeError("Unable to find {} string.".format(attr))


setup(
    name=find_attr('__program__', 'viman/__init__.py'),
    version=find_attr('__version__', 'viman/__init__.py'),
    description='The vim plugin manager based on git and pathogen!',
    long_description=LONG_DESCRIPTION,
    url='https://github.com/Shylock-Hg/viman.git',
    author='Shylock Hg',
    author_email='tcath2s@gmail.com',
    license='GPL-3.0',
    packages=['viman'],
    install_requires=['pyyaml'],
    python_requires='>=3.3',
    entry_points={
        'console_scripts': ['viman=viman.vimanApp:main'],
        },
    zip_safe=False)



#! /usr/bin/env python3

'''
@brief vimanArgParser the submodule of command line parser
@note 'viman <operation> [options...] [targets..]'
'''

import getopt
import sys
import errno
import copy

class vimanOperations():
    # (short,long,des)
    operations = {'sync': ('S', 'sync', 'Synchronize vim plugin package!'),
                  'remove': ('R', 'remove', 'Remove vim plugin package!'),
                  'upgrade': ('U', 'upgrade', 'Upgrade vim plugin package!'),
                  'query': ('Q', 'query', 'Query vim plugin package!'),
                  'version': ('V', 'version', 'Show version of application!'),
                  'help': ('h', 'help', 'Print help information!')}

    @staticmethod
    def getKey4Operation(operation:str):
        '''
        @brief get key by operation name(such as 'Q','sync')
        @param operation name
        @retval key of operations
        '''
        for key_value in vimanOperations.operations.items():
            if operation in key_value[1]:
                return key_value[0]

        return None

    @staticmethod
    def getOperationsShort():
        return [operation[0] for operation in
                vimanOperations.operations.values()
                if not '' == operation[0]]

    @staticmethod
    def getOperationsLong():
        return [operation[1] for operation in
                vimanOperations.operations.values()
                if not '' == operation[1]]

    @staticmethod
    def getOperations():
        return [ops[0] for ops in list(vimanOperations.operations.values())] +\
               [ops[1] for ops in list(vimanOperations.operations.values())]


class vimanOptions():

    # (short,long,operations,des)
    options = {
        'sysupgrade': (
            'u', 'sysupgrade', 'S', 'Upgrade all vim plugin packages!'),
        'file': (
            'f', 'file', 'SRU', 'Specify plugins by yml file!'),
        'recursive': (
            'r', 'recursive', 'U',
            'Recursive upgrade vim plugin packages!'),
        'current': (
            'c', 'current', 'S',
            'Specify a current yml context string'),
        'name': (
            'n', 'name', 'R',
            'Specify the name of plugins to remove!')}

    @staticmethod
    def getKey4Option(option:str):
        '''
        @brief get key by option name(such as 'u','sysupgrade')
        @param option name
        @retval key of options
        '''
        for key_value in vimanOptions.options.items():
            if option in key_value[1]:
                return key_value[0]

    @staticmethod
    def getOptionShort():
        return [option[0] for option in
                vimanOptions.options.values() if not '' == option[0]]

    @staticmethod
    def getOptionLong():
        return [option[1] for option in vimanOptions.options.values()
                if not '' == option[1]]

    @staticmethod
    def getOptions():
        return [opt[0] for opt in list(vimanOptions.options.values())] +\
                [opt[1] for opt in list(vimanOptions.options.values())]


class vimanArgParser():
    '''
    @brief viman command line arguments parser
    @note viman command line synopsis `viman <operation> [options] [targets]`
    '''
    def __init__(self, args):
        '''
        @brief parse command line arguments
        @param args sys.argv[1:]
        @retval generated object
        '''
        options_short = \
             vimanOperations.getOperationsShort() +\
             vimanOptions.getOptionShort()
        # print(options_short)
        options_short = ''.join(options_short)
        options_long = \
                vimanOperations.getOperationsLong() +\
                vimanOptions.getOptionLong()
        # print(options_long)

        opts, rest = getopt.getopt(args, options_short, options_long)

        self.targets = rest  # targets git repository url

        opts = map(lambda opt: opt[0].replace('-', ''), opts)
        # print(list(opts))

        operations = vimanOperations.getOperations()
        options = vimanOptions.getOptions()

        opts_operations = list(filter(
            lambda opt: opt in operations,
            copy.deepcopy(opts)))
        # print(opts_operations)
        opts_options = list(filter(
            lambda opt: opt in options,
            copy.deepcopy(opts)))

        # check operations
        if not 1 == len(opts_operations):
            print(
                    'error: only one operation may be used at a time!',
                    file=sys.stderr)
            sys.exit(errno.EINVAL)

        # check options
        def check_option(option, operation):
            key_option = vimanOptions.getKey4Option(option)
            key_operation = vimanOperations.getKey4Operation(operation)
            if not vimanOperations.operations[key_operation][0] in \
                    vimanOptions.options[key_option][2]:
                print('error:invalid option `{}` for operation `{}`'.format(
                    option,
                    vimanOperations.operations[key_operation][0]),
                    file=sys.stderr)
                sys.exit(errno.EINVAL)

        for opt in opts_options:
            check_option(opt, opts_operations[0])

        if len(opts_operations[0]) > 1:
            # operations
            self.operations = vimanOperations.operations[opts_operations[0]][0]
        else:
            self.operations = opts_operations[0]

        opts_options_short = filter(
                lambda o: 1 == len(o),
                copy.deepcopy(opts_options))
        opts_options_long = filter(
                lambda o: 1 < len(o),
                copy.deepcopy(opts_options))
        opts_options_long = map(
                lambda l: vimanOptions.operations[ops][0],
                opts_options_long)
        opts_options = list(opts_options_short) + list(opts_options_long)
        self.options = opts_options  # options

    @staticmethod
    def printUsage():
        '''
        @brief print usage of application
        '''
        print('\nSynopsis of viman: ')
        print('viman <operation> [options...] [targets...]')
        for ops in vimanOperations.operations.values():
            print('\n\tOperation `-{},--{}` for {}'.format(
                ops[0], ops[1], ops[2]))
            for opt in vimanOptions.options.values():
                if ops[0] in opt[2]:
                    print('\t\tOption `-{},--{}` for {}'.format(
                        opt[0], opt[1], opt[3]))


def _test():
    parser = vimanArgParser(sys.argv[1:])
    print(parser.operations)
    print(parser.options)
    print(parser.targets)

if '__main__' == __name__:
    _test()
